import { providers } from 'ethers';
import { Cache } from '@cerc-io/cache';
import { EthClient as EthClientInterface, EthFullBlock, EthFullTransaction } from '@cerc-io/util';
export interface Config {
    cache: Cache | undefined;
    rpcEndpoint: string;
}
interface Vars {
    blockHash?: string;
    blockNumber?: string;
    contract?: string;
    slot?: string;
    addresses?: string[];
    topics?: string[][];
    fromBlock?: number;
    toBlock?: number;
}
export declare class EthClient implements EthClientInterface {
    _provider: providers.JsonRpcProvider;
    _cache: Cache | undefined;
    constructor(config: Config);
    getStorageAt({ blockHash, contract, slot }: {
        blockHash: string;
        contract: string;
        slot: string;
    }): Promise<{
        value: string;
        proof: {
            data: string;
        };
    }>;
    getBlockWithTransactions({ blockNumber, blockHash }: {
        blockNumber?: number;
        blockHash?: string;
    }): Promise<any>;
    getBlocks({ blockNumber, blockHash }: {
        blockNumber?: number;
        blockHash?: string;
    }): Promise<any>;
    getFullBlocks({ blockNumber, blockHash }: {
        blockNumber?: number;
        blockHash?: string;
    }): Promise<Array<EthFullBlock | null>>;
    getFullTransaction(txHash: string): Promise<EthFullTransaction>;
    getBlockByHash(blockHash?: string): Promise<any>;
    getLogs(vars: {
        blockHash: string;
        blockNumber: string;
        addresses?: string[];
        topics?: string[][];
    }): Promise<any>;
    getLogsForBlockRange(vars: {
        fromBlock?: number;
        toBlock?: number;
        addresses?: string[];
        topics?: string[][];
    }): Promise<any>;
    _getLogs(vars: {
        blockHash?: string;
        fromBlock?: number;
        toBlock?: number;
        addresses?: string[];
        topics?: string[][];
    }): Promise<any>;
    _getCachedOrFetch<Result>(queryName: string, vars: Vars, fetch: () => Promise<Result>): Promise<Result>;
    _handleGetBlockErrors(err: any): Array<null>;
}
export {};
