"use strict";
//
// Copyright 2021 Vulcanize, Inc.
//
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (g && (g = 0, op[0] && (_ = 0)), _) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EthClient = void 0;
var assert_1 = __importDefault(require("assert"));
var ethers_1 = require("ethers");
var util_1 = require("@cerc-io/util");
var ipld_eth_client_1 = require("@cerc-io/ipld-eth-client");
var EthClient = /** @class */ (function () {
    function EthClient(config) {
        var rpcEndpoint = config.rpcEndpoint, cache = config.cache;
        (0, assert_1.default)(rpcEndpoint, 'Missing RPC endpoint');
        this._provider = new util_1.MonitoredStaticJsonRpcProvider({
            url: rpcEndpoint,
            allowGzip: true
        });
        this._cache = cache;
    }
    EthClient.prototype.getStorageAt = function (_a) {
        var blockHash = _a.blockHash, contract = _a.contract, slot = _a.slot;
        return __awaiter(this, void 0, void 0, function () {
            var value;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        slot = "0x".concat((0, ipld_eth_client_1.padKey)(slot));
                        console.time("time:eth-client#getStorageAt-".concat(JSON.stringify({ blockHash: blockHash, contract: contract, slot: slot })));
                        return [4 /*yield*/, this._getCachedOrFetch('getStorageAt', { blockHash: blockHash, contract: contract, slot: slot }, function () { return __awaiter(_this, void 0, void 0, function () {
                                return __generator(this, function (_a) {
                                    // TODO: Check if blockHash works with Lotus RPC
                                    return [2 /*return*/, this._provider.getStorageAt(contract, slot, blockHash)];
                                });
                            }); })];
                    case 1:
                        value = _b.sent();
                        console.timeEnd("time:eth-client#getStorageAt-".concat(JSON.stringify({ blockHash: blockHash, contract: contract, slot: slot })));
                        return [2 /*return*/, {
                                value: value,
                                proof: {
                                    // TODO: Return proof with cid and ipldBlock
                                    // To match getStorageAt method of ipld-eth-client which returns proof along with value.
                                    data: JSON.stringify(null)
                                }
                            }];
                }
            });
        });
    };
    EthClient.prototype.getBlockWithTransactions = function (_a) {
        var blockNumber = _a.blockNumber, blockHash = _a.blockHash;
        return __awaiter(this, void 0, void 0, function () {
            var blockHashOrBlockNumber, nodes, result, err_1;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        blockHashOrBlockNumber = blockHash !== null && blockHash !== void 0 ? blockHash : blockNumber;
                        (0, assert_1.default)(blockHashOrBlockNumber);
                        console.time("time:eth-client#getBlockWithTransactions-".concat(JSON.stringify({ blockNumber: blockNumber, blockHash: blockHash })));
                        nodes = [];
                        _b.label = 1;
                    case 1:
                        _b.trys.push([1, 3, 4, 5]);
                        return [4 /*yield*/, this._provider.getBlockWithTransactions(blockHashOrBlockNumber)];
                    case 2:
                        result = _b.sent();
                        nodes = [
                            {
                                blockNumber: result.number.toString(),
                                blockHash: result.hash,
                                parentHash: result.parentHash,
                                timestamp: result.timestamp.toString(),
                                ethTransactionCidsByHeaderId: {
                                    nodes: result.transactions.map(function (transaction) { return ({
                                        txHash: transaction.hash,
                                        // Transactions with block should be of type TransactionReceipt
                                        index: transaction.transactionIndex,
                                        src: transaction.from,
                                        dst: transaction.to
                                    }); })
                                }
                            }
                        ];
                        return [3 /*break*/, 5];
                    case 3:
                        err_1 = _b.sent();
                        return [2 /*return*/, this._handleGetBlockErrors(err_1)];
                    case 4:
                        console.timeEnd("time:eth-client#getBlockWithTransactions-".concat(JSON.stringify({ blockNumber: blockNumber, blockHash: blockHash })));
                        return [7 /*endfinally*/];
                    case 5: return [2 /*return*/, {
                            allEthHeaderCids: {
                                nodes: nodes
                            }
                        }];
                }
            });
        });
    };
    EthClient.prototype.getBlocks = function (_a) {
        var blockNumber = _a.blockNumber, blockHash = _a.blockHash;
        return __awaiter(this, void 0, void 0, function () {
            var blockNumberHex, blockHashOrBlockNumber, nodes, rawBlock, block, err_2;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        blockNumberHex = blockNumber ? ethers_1.utils.hexValue(blockNumber) : undefined;
                        blockHashOrBlockNumber = blockHash !== null && blockHash !== void 0 ? blockHash : blockNumberHex;
                        (0, assert_1.default)(blockHashOrBlockNumber);
                        nodes = [];
                        console.time("time:eth-client#getBlocks-".concat(JSON.stringify({ blockNumber: blockNumber, blockHash: blockHash })));
                        _b.label = 1;
                    case 1:
                        _b.trys.push([1, 3, 4, 5]);
                        return [4 /*yield*/, this._provider.send(blockHash ? 'eth_getBlockByHash' : 'eth_getBlockByNumber', [blockHashOrBlockNumber, false])];
                    case 2:
                        rawBlock = _b.sent();
                        if (rawBlock) {
                            block = this._provider.formatter.block(rawBlock);
                            nodes = [
                                {
                                    blockNumber: block.number.toString(),
                                    blockHash: block.hash,
                                    parentHash: block.parentHash,
                                    timestamp: block.timestamp.toString(),
                                    stateRoot: this._provider.formatter.hash(rawBlock.stateRoot),
                                    td: this._provider.formatter.bigNumber(rawBlock.totalDifficulty).toString(),
                                    txRoot: this._provider.formatter.hash(rawBlock.transactionsRoot),
                                    receiptRoot: this._provider.formatter.hash(rawBlock.receiptsRoot)
                                }
                            ];
                        }
                        return [3 /*break*/, 5];
                    case 3:
                        err_2 = _b.sent();
                        return [2 /*return*/, this._handleGetBlockErrors(err_2)];
                    case 4:
                        console.timeEnd("time:eth-client#getBlocks-".concat(JSON.stringify({ blockNumber: blockNumber, blockHash: blockHash })));
                        return [7 /*endfinally*/];
                    case 5: return [2 /*return*/, {
                            allEthHeaderCids: {
                                nodes: nodes
                            }
                        }];
                }
            });
        });
    };
    EthClient.prototype.getFullBlocks = function (_a) {
        var blockNumber = _a.blockNumber, blockHash = _a.blockHash;
        return __awaiter(this, void 0, void 0, function () {
            var blockNumberHex, blockHashOrBlockNumber, rawBlock, header, rlpData, err_3;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        blockNumberHex = blockNumber ? ethers_1.utils.hexValue(blockNumber) : undefined;
                        blockHashOrBlockNumber = blockHash !== null && blockHash !== void 0 ? blockHash : blockNumberHex;
                        (0, assert_1.default)(blockHashOrBlockNumber);
                        console.time("time:eth-client#getFullBlocks-".concat(JSON.stringify({ blockNumber: blockNumber, blockHash: blockHash })));
                        _b.label = 1;
                    case 1:
                        _b.trys.push([1, 3, 4, 5]);
                        return [4 /*yield*/, this._provider.send(blockHash ? 'eth_getBlockByHash' : 'eth_getBlockByNumber', [blockHashOrBlockNumber, false])];
                    case 2:
                        rawBlock = _b.sent();
                        if (rawBlock) {
                            header = {
                                Parent: rawBlock.parentHash,
                                UnclesDigest: rawBlock.sha3Uncles,
                                Beneficiary: rawBlock.miner,
                                StateRoot: rawBlock.stateRoot,
                                TxRoot: rawBlock.transactionsRoot,
                                RctRoot: rawBlock.receiptsRoot,
                                Bloom: rawBlock.logsBloom,
                                Difficulty: BigInt(rawBlock.difficulty),
                                Number: BigInt(rawBlock.number),
                                GasLimit: BigInt(rawBlock.gasLimit),
                                GasUsed: BigInt(rawBlock.gasUsed),
                                Time: Number(rawBlock.timestamp),
                                Extra: rawBlock.extraData,
                                MixDigest: rawBlock.mixHash,
                                Nonce: BigInt(rawBlock.nonce),
                                BaseFee: rawBlock.baseFeePerGas && BigInt(rawBlock.baseFeePerGas)
                            };
                            rlpData = (0, util_1.encodeHeader)(header);
                            return [2 /*return*/, [{
                                        blockNumber: this._provider.formatter.number(rawBlock.number).toString(),
                                        blockHash: this._provider.formatter.hash(rawBlock.hash),
                                        parentHash: this._provider.formatter.hash(rawBlock.parentHash),
                                        timestamp: this._provider.formatter.number(rawBlock.timestamp).toString(),
                                        stateRoot: this._provider.formatter.hash(rawBlock.stateRoot),
                                        td: this._provider.formatter.bigNumber(rawBlock.totalDifficulty).toString(),
                                        txRoot: this._provider.formatter.hash(rawBlock.transactionsRoot),
                                        receiptRoot: this._provider.formatter.hash(rawBlock.receiptsRoot),
                                        uncleRoot: this._provider.formatter.hash(rawBlock.sha3Uncles),
                                        bloom: (0, util_1.escapeHexString)(this._provider.formatter.hex(rawBlock.logsBloom)),
                                        size: this._provider.formatter.number(rawBlock.size).toString(),
                                        blockByMhKey: {
                                            data: (0, util_1.escapeHexString)(rlpData)
                                        }
                                    }]];
                        }
                        return [3 /*break*/, 5];
                    case 3:
                        err_3 = _b.sent();
                        return [2 /*return*/, this._handleGetBlockErrors(err_3)];
                    case 4:
                        console.timeEnd("time:eth-client#getFullBlocks-".concat(JSON.stringify({ blockNumber: blockNumber, blockHash: blockHash })));
                        return [7 /*endfinally*/];
                    case 5: return [2 /*return*/, []];
                }
            });
        });
    };
    EthClient.prototype.getFullTransaction = function (txHash) {
        return __awaiter(this, void 0, void 0, function () {
            var tx;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        console.time("time:eth-client#getFullTransaction-".concat(JSON.stringify({ txHash: txHash })));
                        return [4 /*yield*/, this._provider.getTransaction(txHash)];
                    case 1:
                        tx = _a.sent();
                        console.timeEnd("time:eth-client#getFullTransaction-".concat(JSON.stringify({ txHash: txHash })));
                        return [2 /*return*/, {
                                ethTransactionCidByTxHash: {
                                    txHash: tx.hash,
                                    index: tx.transactionIndex,
                                    src: tx.from,
                                    dst: tx.to
                                },
                                data: tx
                            }];
                }
            });
        });
    };
    EthClient.prototype.getBlockByHash = function (blockHash) {
        return __awaiter(this, void 0, void 0, function () {
            var blockTag, block;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        blockTag = blockHash !== null && blockHash !== void 0 ? blockHash : 'latest';
                        console.time("time:eth-client#getBlockByHash-".concat(blockHash));
                        return [4 /*yield*/, this._provider.getBlock(blockTag)];
                    case 1:
                        block = _a.sent();
                        console.timeEnd("time:eth-client#getBlockByHash-".concat(blockHash));
                        return [2 /*return*/, {
                                block: {
                                    number: block.number,
                                    hash: block.hash,
                                    parent: {
                                        hash: block.parentHash
                                    },
                                    timestamp: block.timestamp
                                }
                            }];
                }
            });
        });
    };
    EthClient.prototype.getLogs = function (vars) {
        return __awaiter(this, void 0, void 0, function () {
            var result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        console.time("time:eth-client#getLogs-".concat(JSON.stringify(vars)));
                        return [4 /*yield*/, this._getLogs({
                                blockHash: vars.blockHash,
                                addresses: vars.addresses,
                                topics: vars.topics
                            })];
                    case 1:
                        result = _a.sent();
                        console.timeEnd("time:eth-client#getLogs-".concat(JSON.stringify(vars)));
                        return [2 /*return*/, result];
                }
            });
        });
    };
    EthClient.prototype.getLogsForBlockRange = function (vars) {
        return __awaiter(this, void 0, void 0, function () {
            var result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        console.time("time:eth-client#getLogsForBlockRange-".concat(JSON.stringify(vars)));
                        return [4 /*yield*/, this._getLogs({
                                fromBlock: Number(vars.fromBlock),
                                toBlock: Number(vars.toBlock),
                                addresses: vars.addresses,
                                topics: vars.topics
                            })];
                    case 1:
                        result = _a.sent();
                        console.timeEnd("time:eth-client#getLogsForBlockRange-".concat(JSON.stringify(vars)));
                        return [2 /*return*/, result];
                }
            });
        });
    };
    // TODO: Implement return type
    EthClient.prototype._getLogs = function (vars) {
        return __awaiter(this, void 0, void 0, function () {
            var blockHash, fromBlock, toBlock, _a, addresses, topics, result, txHashesSet, txReceipts, txReceiptMap;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        blockHash = vars.blockHash, fromBlock = vars.fromBlock, toBlock = vars.toBlock, _a = vars.addresses, addresses = _a === void 0 ? [] : _a, topics = vars.topics;
                        return [4 /*yield*/, this._getCachedOrFetch('getLogs', vars, function () { return __awaiter(_this, void 0, void 0, function () {
                                var ethLogs, logs;
                                return __generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4 /*yield*/, this._provider.send('eth_getLogs', [{
                                                    address: addresses.map(function (address) { return address.toLowerCase(); }),
                                                    fromBlock: fromBlock && ethers_1.utils.hexValue(fromBlock),
                                                    toBlock: toBlock && ethers_1.utils.hexValue(toBlock),
                                                    blockHash: blockHash,
                                                    topics: topics
                                                }])];
                                        case 1:
                                            ethLogs = _a.sent();
                                            logs = ethers_1.providers.Formatter.arrayOf(this._provider.formatter.filterLog.bind(this._provider.formatter))(ethLogs);
                                            return [2 /*return*/, logs.map(function (log) {
                                                    log.address = log.address.toLowerCase();
                                                    return log;
                                                })];
                                    }
                                });
                            }); })];
                    case 1:
                        result = _b.sent();
                        txHashesSet = result.reduce(function (acc, log) {
                            acc.add(log.transactionHash);
                            return acc;
                        }, new Set());
                        return [4 /*yield*/, Promise.all(Array.from(txHashesSet).map(function (txHash) { return _this._provider.getTransactionReceipt(txHash); }))];
                    case 2:
                        txReceipts = _b.sent();
                        txReceiptMap = txReceipts.reduce(function (acc, txReceipt) {
                            acc.set(txReceipt.transactionHash, txReceipt);
                            return acc;
                        }, new Map());
                        return [2 /*return*/, {
                                logs: result.map(function (log) {
                                    var _a;
                                    return ({
                                        // blockHash required for sorting logs fetched in a block range
                                        blockHash: log.blockHash,
                                        account: {
                                            address: log.address
                                        },
                                        transaction: {
                                            hash: log.transactionHash
                                        },
                                        topics: log.topics,
                                        data: log.data,
                                        index: log.logIndex,
                                        status: (_a = txReceiptMap.get(log.transactionHash)) === null || _a === void 0 ? void 0 : _a.status
                                    });
                                })
                            }];
                }
            });
        });
    };
    EthClient.prototype._getCachedOrFetch = function (queryName, vars, fetch) {
        return __awaiter(this, void 0, void 0, function () {
            var keyObj, _a, value, found, result;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        keyObj = {
                            queryName: queryName,
                            vars: vars
                        };
                        if (!this._cache) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._cache.get(keyObj)];
                    case 1:
                        _a = (_b.sent()) || [undefined, false], value = _a[0], found = _a[1];
                        if (found) {
                            return [2 /*return*/, value];
                        }
                        _b.label = 2;
                    case 2: return [4 /*yield*/, fetch()];
                    case 3:
                        result = _b.sent();
                        if (!this._cache) return [3 /*break*/, 5];
                        return [4 /*yield*/, this._cache.put(keyObj, result)];
                    case 4:
                        _b.sent();
                        _b.label = 5;
                    case 5: return [2 /*return*/, result];
                }
            });
        });
    };
    EthClient.prototype._handleGetBlockErrors = function (err) {
        if (err.code === ethers_1.errors.SERVER_ERROR && err.error) {
            // Check null block error and return null array
            if (err.error.message === util_1.NULL_BLOCK_ERROR) {
                return [null];
            }
            // Check and ignore future block error
            if (err.error.message === util_1.FUTURE_BLOCK_ERROR) {
                return [];
            }
        }
        throw err;
    };
    return EthClient;
}());
exports.EthClient = EthClient;
//# sourceMappingURL=eth-client.js.map